///////////////////////////////////////////////////////////////////////////////////
// Gamestudio test level
///////////////////////////////////////////////////////////////////////////////////
path "..\\template";	// Path to templates subdirectory

include <entstrings.wdl>;
include <movement.wdl>;	// libraries of WDL functions
include <messages.wdl>;
include <menu.wdl>;		// menu must be included BEFORE doors and weapons
include <particle.wdl>; // remove when you need no particles
include <doors.wdl>;		// remove when you need no doors
include <actors.wdl>;   // remove when you need no actors
include <weapons.wdl>;  // remove when you need no weapons
include <war.wdl>;      // remove when you need no fighting

///////////////////////////////////////////////////////////////////////////////////
// define all file names in angular brackets for including them into the resource
string office_wmb = <office.wmb>;
string fish_mdl = <fish.mdl>;
string shield_mdl = <shield.mdl>;
string mace_mdl = <mace.mdl>;
string warlock_mdl = <warlock.mdl>;
string staff_mdl = <staff.mdl>;
string blitz_pcx = <blitz.pcx>;

///////////////////////////////////////////////////////////////////////////////////
// The engine starts in the resolution given by the variable
// VIDEO_MODE. It is possible to switch the resolution during
// gameplay using the SWITCH_VIDEO instruction.

var video_mode = 6; 		// 640x480;
var video_depth = 16;	// D3D mode
//var d3d_mipmapping = 1; // old method
var media_layer = 0.5;	// play movies behind all panels
//var dplay_unreliable = -1;	// always use TCP/IP protocol
//var d3d_autotransparency = on;

var d3d_lockable = 1;

/////////////////////////////////////////////////////////////////
// define a splash screen with the required A4/A5 logo
bmap splashmap = <logodark.pcx>; // the default A5 logo in templates
panel splashscreen { bmap = splashmap; flags = refresh,d3d; }

////////////////////////////////////////////////////////////////////////////
// function prototypes
function player_client();
function start_song();

////////////////////////////////////////////////////////////////////////////
// After engine start, the MAIN function is executed, so the 2-D engine
// can display pictures, logos, or AVI animations. To load
// a level, perform a LEVEL_LOAD instruction.

function main()
{
// set some global values for the test level, like the walking
// animation adaption
	anim_walk_dist = 2; // for the new guards
	anim_run_dist = 3;
	walk_or_run = 8;
	camera_dist.z = -50;

	fps_max = 80;
	warn_level = 2;	// announce bad texture sizes and bad wdl code
	tex_share = on;	// map entities share their textures

// center the splash screen for non-640x480 resolutions
	splashscreen.pos_x = (screen_size.x - bmap_width(splashmap))/2;
	splashscreen.pos_y = (screen_size.y - bmap_height(splashmap))/2;
// set it visible
//	splashscreen.visible = on;

// wait 3(!) frames for triple buffering, until it is flipped to the foreground
	wait(3);

// now load the level
	level_load(office_wmb);
// wait the required second, then switch the splashscreen off.
	sleep(1);

  	splashscreen.visible = off;
	bmap_purge(splashmap);	// remove logo bitmap from video memory

// load some global variables, like sound volume or changed strings
//	load_status();
// initalize the sky and background

// initialize lens flares in D3D mode (only if edition supports flares)

// comment this in if you really want to play a midi file
//	start_song();

// create the player entity on the client
	if (connection & CONNECT_CLIENT)
	{
		client_move();	// enable multiuser mode

	// wait until the level is loaded, and the connection is established
		while (connection == 0) { wait(1); }

	// create a client entity at a random place
		randomize();
		temp.X = 300 + random(100);
		temp.Y = -400+ random(100);
		temp.Z = 50;
		player = ent_create(warlock_mdl,temp,player_client);
	// the guard is created, and its action is running on the server.
	// Therefore we have to initialize him on the client separately -
	// the following flags are required for the camera
		player._MOVEMODE = _MODE_WALKING;
		player.__BOB = ON;
	}
// display the server's IP address
	if (connection & CONNECT_SERVER)
	{
		str_cpy(temp_str,server_name);
		str_cat(temp_str," ");
		str_cat(temp_str,server_IP);
		scroll_message(temp_str);
	}
	while (player == null) {wait (1);}
	player._armor = 10;	
	breakpoint_on(player._armor < 10);
}


// play a song if volume was set at beginning
function start_song()
{
	wait(4);	// wait until LOAD_INFO has loaded the last volume setting
	if (midi_vol > 0) {
//		media_play("ribanna.mid",NULL,100);
	}
}

////////////////////////////////////////////////////////////////////////

function zoom_in() {
	camera.arc = max(camera.arc-5,10);
}

function zoom_out() {
	camera.arc = min(camera.arc+5,120);
}

VIEW mirror { };

// Code to simulate a floor mirror by generating
// a vertically flipped camera view from below through the floor
action mirror_vertical
{
// initialize camera portal for realtime reflections
//ifdef CAPS_MIRROR;
	camera.portal = mirror;
	mirror.noshadow = on;	// suppress shadows in the mirror
	mirror.portalclip = on;	// clip at portal plane
	while (1) {
		proc_late();	// place it at the end of the function list - the camera must be moved before
		mirror.genius = camera.genius;
		mirror.aspect = -camera.aspect;	// flip the image upside down
		mirror.arc = camera.arc;
		mirror.fog = camera.fog;
		mirror.fog_start = camera.fog_start;
		mirror.fog_end = camera.fog_end;
		mirror.x = camera.x;
		mirror.y = camera.y;
		mirror.z = 2*camera.portal_z-camera.z;	// move the camera at its mirror position
		mirror.pan = camera.pan;
		mirror.tilt = -camera.tilt;	// flip the vertical camera angle
		mirror.roll = -camera.roll;
		wait(1);
	}
//endif;
	return;
}

////////////////////////////////////////////////////////////////////////////
// Everything that moves is an ENTITY. Entities may be placed
// into the level, or may be created at runtime. There will be
// 3 types of entities supported:
//
// - Model Entities (MDL files) for animated actors or objects
// - Sprite Entities (PCX files) for animated actors, objects, or explosions
// - Block Entities (WMB files) for doors, platforms, buildings and the like
//
// Each entity has two functions: The main AI action, which initializes it and controls
// its behaviour, and a separate EVENT function, which is triggered on
// certain events, like being hit or being touched.

entity* guard;	// define a pointer, thus the guard can be accessed from outside


action patrol_prog
{
	name_ent2();
	guard = me;
//	watched = me;

// first, set the entity properties
	my.fat = off;
	my.narrow = on;
	my._walkframes = 1;
	my._movemode = _mode_walking;
	my._force = 1.5;

// then, call all standard tasks the entity shall perform
	patrol_path();
	if (my.shadow == off) { drop_shadow(); }

// add an event: touching it with the mouse
	my.string1 = "walking\nguard";
	my.enable_touch = on;
	my.event = handle_touch;

//	my.enable_click = on;
//	my.event = _save;

// attach other entities - AFTER setting the patrol movement action
	ent_create(shield_mdl,nullvector,attach_entity);
	ent_create(mace_mdl,nullvector,attach_entity);

// a non-standard entity function - like disappering
// on pressing the [R] key - may be directly scripted here, like this:
	while (1) {
		if (KEY_R) { remove(ME); }
		wait(1);
	}
}

///////////////////////////////////////////////////////////////////////
// Create a player that carries a staff, has a bright light on to of the
// staff, and can shoot fishes when [7] is pressed

// multiplayer helper function: hide this entity only on the client
// who has created it (as for model attachments, like a staff).
function hide_on_client()
{
	while(1)
   {
// hide this entity in a client's 1st person mode
      if (person_3rd == 0) {
			my.invisible = on; }
		else {
			my.invisible = off; }
		wait(1);
   }
}

// place a bright flare at the vertex number given by my.x
function attach_flare(name,vertex)
{
	you = me;	// store the parent entity
	me = ent_create(name,nullvector,null);
	my.passable = on;
	my.flare = on;
	my.unlit = on;
	my.bright = on;
	my.facing = on;

	proc_client(me,hide_on_client);	// start a client function

   while (you)	// prevent empty pointer error if parent entity was removed
   {
		my.invisible = (you.invisible == on);
   	vec_for_vertex(my.x,you,vertex);
   	wait(1);
   }
	remove(me);
}

// this gives the player a staff with a flare
action staff_prog
{
	proc_client(me,hide_on_client);
	attach_entity();
// create a flare sprite at top of the staff
	attach_flare(blitz_pcx,26);
}

define _FISH,7;
function emit_fish(&pos,&angle);

// this function lets the player emit a fish when his _SIGNAL is set to _FISH
function perform_fish()
{
	while(1)
	{
		if (my._SIGNAL == _FISH)
		{	// client has pressed fish key
			my._SIGNAL = 0;				// reset it
			emit_fish(my.x,my.pan);
		}
		wait(1);
	}
}

// this multiplayer-capable function sets the player's _SIGNAL to _FISH
function send_fish()
{
	if(player)
	{
		player._SIGNAL = _FISH;			// send command to shoot a fish
		send_skill(player._SIGNAL,0);	// make it work in multiplayer mode
	}
}

// main player action
function player_client()
{
	my.fat = off;
	my.narrow = on;		// set narrow hull
	my._force = 0.5;		// he should be not too fast
	my._movemode = _mode_walking;
	my._banking = -0.1;
	my.__jump = on;
	my.__duck = on;
	my.__strafe = on;
	my.__bob = on;
	my.__trigger = on;
	my._health = 100;
 	my._runthreshold = 7;

// let him disappear when the client disconnects
	my.enable_disconnect = on;
	my.event = _actor_connect;

 	player_anim_pack();
 	player_move2();		// new move function with animation blending
	perform_fish();		// make player sensitive for the fish key

// give him a shadow
ifdef caps_shadow;
	my.shadow = on;
ifelse;
	drop_shadow();	// fake shadow on lower editions
endif;

// give him a staff with a light
	ent_create(staff_mdl,nullvector,staff_prog);
}

//resource "warlock.wrs";	// a possible resource
//include "warlock.wdl";

action player_prog
{
	name_ent1();
	if (connection) {	// remove the level placed entity in a multiplayer system
		remove(me);	}	// because each client generates his own player
	else {
		player_client(); }
}

entity* water_wmb;

action waterblock {
	water_wmb = my;
// uv offset
//	while (1) { my.v += time; wait(1); }
	my.uvspeed = on;
	my.v = 1.5;
}



// Use the ent_create instruction to create an entity at a given
// position, running a given action:

var drop_dist[3] = 0,0,0;
function fly_ahead();

function emit_fish(&pos,&angle)
{
// wait until the previous fish is far enough
	while (drop_dist.x != 0) { wait(1); }
	drop_dist.x = 60;
// rotate the fish position in direction of the given angle
	vec_set(temp,drop_dist);
	vec_rotate(temp,angle);
// create the fish 60 quants ahead of the given position
	vec_add(temp,pos);
	you = ent_create(fish_mdl,temp,fly_ahead);
	vec_set(your.pan,angle);
	waitt(8);
	drop_dist.x = 0;
}

// This is the entity's action. An entity can change its type,
// (so a model can become a sprite), can be attached to another
// entity, and can perform some move or AI instructions.

var max_lights = 3;
SOUND bumm = <bumm.wav>;

// sets the vector to a random direction and length (max. range)
function vec_randomize(&vec,range)
{
	vec[0] = random(1) - 0.5;
	vec[1] = random(1) - 0.5;
	vec[2] = random(1) - 0.5;
	vec_normalize(vec,random(range));
}

// fades out a particle
function part_alphafade()
{
	my.alpha -= 2*time;
	if (my.alpha <= 0) { my.lifespan = 0; }
}


function effect_explo()
{
	vec_randomize(temp,10);
	vec_add(my.vel_x,temp);

	my.lifespan = 100;
	my.gravity = 0.5;
	my.bmap = scatter_map;
	my.flare = on;
	my.bright = on;
	my.streak = on;
	my.alpha = 50;
	my.move = on;
	my.function = part_alphafade;
}

function event_ricochet()
{
	ent_playsound(my,bumm,100);
// bounce off the wall
	vec_to_angle(my.pan,bounce);
// emit particles into normal direction
	vec_scale(normal,5);
	effect(effect_explo,1000,my.x,normal);
}

sound whosh = <noise.wav>;

function fly_ahead()
{
	my.event = event_ricochet;
	my.enable_block = on;
	my.enable_entity = on;

	if	(max_lights > 0) {
		my.red = 50;
		my.green = 50;
		my.blue = 200;
		my.lightrange = 200;
		max_lights -= 1;
	}

	var whosh_handle;
	whosh_handle = ent_playloop(my,whosh,100);

	while (1) {
		move_mode = ignore_you;
		var speed[3];
		speed.x = 10 * time;
		ent_move(speed,nullvector);
// tune the entity sound
ifndef SERVER;	// snd_tune does not work on the server because sounds play on the client
//		snd_tune(whosh_handle,(10*total_ticks) % 100,0,0);
endif;
// use the modulo function to cycle colors from 0..255
		my.red = (4*total_ticks) % 256;
		my.green = (3*total_ticks) % 256;
		my.blue = total_ticks % 256;

		wait(1);
	}
}

/////////////////////////////////////////////////////////////////////////
// The following definitions control the sky

var d3d_alphadepth = 32;

/////////////////////////////////////////////////////////////////
// For changing a level, just perform a level_load instruction
function levelchange()
{
	level_load("office2.wmb");
}

// animate an entity by changing its frames
action animate
{
	while(1) {
		my.frame += time;
		if (my.frame > ent_frames(my)) {
			my.frame -= ent_frames(my); }
		wait(1);
	}
}

action animate_once
{
	my.frame = 1;
	while(my.frame < ent_frames(my)) {
		wait(1);
    	my.frame += time;
	}
}

/////////////////////////////////////////////////////////////////
// The following definitions are for the WDFC window composer
// (available only with the professional edition)
// to define the start and exit window of the application.
WINDOW WINSTART
{
	TITLE			"A5 Engine";
	SIZE			480,320;
	MODE			IMAGE;		// STANDARD;
	BG_COLOR		RGB(0,0,0);	// black
	FRAME			FTYP1,0,0,480,320;
//	BUTTON		BUTTON_START,SYS_DEFAULT,"Start",400,288,72,24;
	BUTTON		BUTTON_QUIT,SYS_DEFAULT,"Abort",400,288,72,24;
	TEXT_STDOUT	"Arial",RGB(255,0,0),10,10,460,280;
}

// no exit window at all..
/*WINDOW WINEND
{
	TITLE			"A5 Finished";
	SIZE			540,320;
	MODE	 		STANDARD;
	BG_COLOR		RGB(0,0,0);
	TEXT_STDOUT	"Arial",RGB(255,40,40),10,20,520,270;

	SET FONT		"Arial",RGB(0,255,255);
	TEXT			"Any key to exit",10,270;
}*/

/////////////////////////////////////////////////////////////////
include <debug.wdl>;

on_2 = levelchange;
on_3 = zoom_in;
on_4 = zoom_out;
on_6 = snowfall;
on_7 = send_fish;